﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.IO;
using System.Xml;
using Microsoft.Win32;
using ICSharpCode.SharpZipLib.Zip;
using System.Diagnostics;
using Utils.Misc;


namespace PhoenixUpdate.Class
{
    /// <summary>
    /// EventHandler for an update notification
    /// </summary>
    /// <param name="sender"></param>
    /// <param name="e"></param>    
    public delegate void OnUpdateChangedHandler(object sender, XmlUpdateEventArgs e);
    /// <summary>
    /// Manage the update XML file
    /// </summary>
    public class XmlUpdateManager
    {
        #region Variables
        #region static readonly
        /// <summary>
        /// Name of the root node of the XML
        /// </summary>
        private static readonly string vRootNode = "Phoenix";
        /// <summary>
        /// Name of the downloaded update
        /// </summary>
        private static readonly string vFileLocation = Path.GetTempPath() + DateTime.Now.Ticks.ToString() + ".zip";
        #endregion

        /// <summary>
        /// Path of the XML file
        /// </summary>
        private string vXMLPath = "http://download.tuxfamily.org/phoenixsuite/Update.XML";
        /// <summary>
        /// New version of Phoenix Suite
        /// </summary>
        private int vVersion;
        /// <summary>
        /// Readable version form
        /// </summary>
        private string vVersionTitle;
        /// <summary>
        /// Url wherin is the 
        /// </summary>
        private string vUrl;
        /// <summary>
        /// Date of the release
        /// </summary>
        private DateTime vDate;
        /// <summary>
        /// Indicates whether keep the options file
        /// </summary>
        private bool vKeepOptions = false;
        /// <summary>
        /// Information about the update
        /// </summary>
        private string vText = "";
        /// <summary>
        /// Uri of the ftp server
        /// </summary>
        private string vFtpUri = "";
        /// <summary>
        /// Download object
        /// </summary>
        WebClient vWebClient = new WebClient();
        /// <summary>
        /// The list of executables to launch
        /// </summary>
        private Executable[] vExecutables = null;
        #endregion

        #region Constructor
        /// <summary>
        /// Constructor
        /// </summary>
        public XmlUpdateManager() { }
        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="_XMLPath">Indicates the url of the distant XML file</param>
        public XmlUpdateManager(string _XMLPath) { vXMLPath = _XMLPath; }
        #endregion

        #region Methods
        /// <summary>
        /// Fill the object with the information of the XML File
        /// </summary>
        /// <exception cref="Exception">If an error occures</exception>
        public void Fill()
        {
            try
            {
                WebClient wc = new WebClient();
                Stream s = wc.OpenRead(this.vXMLPath);
                XmlDocument vXmlDocument = new XmlDocument();
                vXmlDocument.Load(s);

                this.vVersion = Int32.Parse(vXmlDocument[vRootNode]["Version"].InnerText);
                this.vVersionTitle = vXmlDocument[vRootNode]["VersionTitle"].InnerText;
                this.vUrl = vXmlDocument[vRootNode]["Url"].InnerText;
                string vvDate = vXmlDocument[vRootNode]["Date"].InnerText;
                this.vDate = DateTime.Parse(vvDate);
                this.vKeepOptions = Boolean.Parse(vXmlDocument[vRootNode]["KeepOptions"].InnerText.ToLower());
                this.vText = vXmlDocument[vRootNode]["Text"].InnerText;

                #region Executables
                XmlNodeList nl = vXmlDocument[vRootNode].SelectNodes("//Executable");
                Executable e;
                this.vExecutables = new Executable[nl.Count];
                for (int i = 0; i < vExecutables.Length; i++)
                {
                    e.ExecutableTitle = nl[i].ChildNodes[0].InnerText;
                    e.ExecutablePath = nl[i].ChildNodes[1].InnerText;
                    this.vExecutables[i] = e;
                }
                #endregion

                s.Close();
                WebClient.Dispose();
            }
            catch (Exception ex)
            {
                throw ex;
            }
        }
        /// <summary>
        /// Start the update of Phoenix Suite
        /// </summary>
        public void Update()
        {
            try
            {
                if (this.vUrl == "" || this.vUrl == null)
                    throw new ArgumentException("L'addresse de téléchargement n'existe pas");

                #region Download the file
                vWebClient.DownloadProgressChanged += new DownloadProgressChangedEventHandler(Event_DownloadProgressChanged);
                vWebClient.DownloadFileCompleted += new System.ComponentModel.AsyncCompletedEventHandler(Event_DownloadFileCompleted);
                Uri uri = new Uri(this.vUrl);
                vWebClient.DownloadFileAsync(uri, vFileLocation);
                #endregion
            }
            catch (Exception ex)
            {
                throw ex;
            }
        }
        /// <summary>
        /// Save the data from the object to the distant Xml file
        /// </summary>
        public void Save()
        {
            try
            {
                if (this.FtpDirectory == null || this.FtpDirectory == ""
                    || this.NetworkCredential == null
                    || this.FtpUri == null || this.FtpUri == "")
                    throw new Exception("Options de connexion au serveur distant non trouvées");


                FileInfo fi = new FileInfo(Path.GetTempPath() + "Update.XML");
                XmlTextWriter wr = new XmlTextWriter(fi.FullName, Encoding.UTF8);

                wr.Formatting = Formatting.Indented;
                wr.WriteStartDocument(true);
                wr.WriteStartElement(XmlUpdateManager.vRootNode);
                wr.WriteElementString("Version", this.vVersion.ToString());
                wr.WriteElementString("VersionTitle", this.vVersionTitle);
                wr.WriteElementString("Url", this.vUrl);
                wr.WriteElementString("Date", this.vDate.Day + "/" + this.vDate.Month + "/" + this.vDate.Year);
                wr.WriteElementString("KeepOptions", this.vKeepOptions.ToString());
                wr.WriteElementString("Text", this.Text);
                wr.WriteStartElement("PostInstall");
                foreach (Executable e in this.Executables)
                {
                    wr.WriteStartElement("Executable");
                    wr.WriteElementString("Name", e.ExecutableTitle);
                    wr.WriteElementString("ExecutableRelativePath", e.ExecutablePath);
                    wr.WriteEndElement();
                }
                wr.WriteEndElement();
                wr.WriteEndElement();
                wr.WriteEndDocument();
                wr.Flush();
                wr.Close();

                vWebClient = new WebClient();
                vWebClient.Credentials = this.NetworkCredential;
                string fp = this.FtpUri + this.FtpDirectory + "/" + fi.Name;
                vWebClient.UploadFile(fp, fi.FullName);
                fi.Delete();
            }
            catch (Exception ex)
            {
                throw ex;
            }
        }
        #endregion

        #region Properties
        /// <summary>
        /// Gets or sets the numeric weight of the latest version
        /// </summary>
        public int Version { get { return this.vVersion; } set { this.vVersion = value; } }
        /// <summary>
        /// Gets or sets the title of the latest version
        /// </summary>
        public string VersionTitle { get { return this.vVersionTitle; } set { this.vVersionTitle = value; } }
        /// <summary>
        /// Gets or sets the URL wherin download the latest version
        /// </summary>
        public string Url { get { return this.vUrl; } set { this.vUrl = value; } }
        /// <summary>
        /// Gets or sets the date of the latest release
        /// </summary>
        public DateTime Date { get { return this.vDate; } set { this.vDate = value; } }
        /// <summary>
        /// Gets or setsthe information about this update
        /// </summary>
        public string Text { get { return this.vText; } set { this.vText = value; } }
        /// <summary>
        /// Gets the webclient for the download
        /// </summary>
        public WebClient WebClient { get { return this.vWebClient; } }
        /// <summary>
        /// Gets or sets the address of the ftp site wherin is the Xml update file
        /// </summary>
        public string FtpUri
        {
            get { return vFtpUri; }
            set
            {
                if (!vFtpUri.Contains("ftp://"))
                    vFtpUri = "ftp://" + vFtpUri;
                vFtpUri = value;
            }
        }
        /// <summary>
        /// Gets or sets the directory wherin is the Xml file
        /// </summary>
        public string FtpDirectory { get; set; }
        /// <summary>
        /// Gets or sets the creditentials of the ftp site
        /// </summary>
        public NetworkCredential NetworkCredential { get; set; }
        /// <summary>
        /// Gets or sets the list of executables to launch after the update
        /// </summary>
        public Executable[] Executables { get { return vExecutables; } set { vExecutables = value; } }
        /// <summary>
        /// Gets the path of Phoenix Suite
        /// </summary>
        public string PhoenixPath
        {
            get
            {
                RegistryKey RKey = Registry.CurrentUser.OpenSubKey(@"Software\Phoenix\");
                return (string)RKey.GetValue("Path");
            }
        }
        #endregion

        #region Events
        /// <summary>
        /// Update information
        /// </summary>
        public event OnUpdateChangedHandler UpdateChanged;
        /// <summary>
        /// Notify an update modification
        /// </summary>
        /// <param name="args"></param>
        protected virtual void OnUpdateChanged(XmlUpdateEventArgs args)
        {
            if (UpdateChanged != null)
                UpdateChanged(this, args);
        }
        /// <summary>
        /// The progress of the download changed
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void Event_DownloadProgressChanged(object sender, DownloadProgressChangedEventArgs e)
        {
            XmlUpdateEventArgs args = new XmlUpdateEventArgs();
            args.DownloadPercentage = e.ProgressPercentage;
            args.IsDownloading = true;
            args.MessageToDisplay = "\r\nTéléchargement de la mise à jour " + this.vVersionTitle;
            OnUpdateChanged(args);
        }
        /// <summary>
        /// The download is over... continue the update
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void Event_DownloadFileCompleted(object sender, System.ComponentModel.AsyncCompletedEventArgs e)
        {
            XmlUpdateEventArgs args = new XmlUpdateEventArgs();
            string AppPath;

            #region Gather information about file to change
            try
            {
                AppPath = this.PhoenixPath;
                args.IsDownloading = false;
                args.IsUpated = false;
                args.MessageToDisplay = "\r\nRépertoire d'installation: " + AppPath;
                OnUpdateChanged(args);
            }
            catch (Exception)
            {
                args.MessageToDisplay = "\r\nImpossible de trouver les fichiers de Phoenix Suite dans la base de registre!";
                args.IsUpated = true;
                OnUpdateChanged(args);
                return;
            }
            #endregion

            #region Unzip the file
            DirectoryInfo di;
            try
            {
                FastZip fz = new FastZip();
                string PathBuffer = vFileLocation.Remove(vFileLocation.Length - 4);
                fz.ExtractZip(vFileLocation, PathBuffer, "");
                di = new DirectoryInfo(PathBuffer);

                args.MessageToDisplay = "\r\nFichiers décompressés.";
                OnUpdateChanged(args);
            }
            catch (Exception)
            {
                args.MessageToDisplay = "\r\nImpossible de décompresser les fichier!"; ;
                args.IsUpated = true;
                OnUpdateChanged(args);
                return;
            }
            #endregion

            #region Copy files

            foreach (FileInfo fi in di.GetFiles())
            {
                try
                {
                    fi.CopyTo(AppPath + @"\" + fi.Name, true);
                    args.MessageToDisplay = "\r\nCopie de " + fi.Name + ".";
                    OnUpdateChanged(args);
                }
                catch (Exception)
                {
                    if (fi.Name != "ICSharpCode.SharpZipLib.dll" &&
                        fi.Name != "PhoenixUpdate.exe")
                    {
                        args.MessageToDisplay = "\r\nImpossible de copier " + fi.Name + "!";
                        args.IsUpated = true;
                        OnUpdateChanged(args);
                    }
                }
            }

            #endregion

            #region Process options
            try
            {
                File.Delete(AppPath + "\\Options.srz");
                args.MessageToDisplay = "\r\nEffacement des options.";

                OnUpdateChanged(args);
            }
            catch (Exception)
            {
                args.MessageToDisplay = "\r\nImpossible de traiter le fichier d'options!";
                args.IsUpated = true;
                OnUpdateChanged(args);
                return;
            }
            #endregion

            #region Finish update
            try
            {
                args.MessageToDisplay = "\r\nMise à jour terminée. Dernières opérations: \r\n";

                foreach (Executable exec in this.Executables)
                {
                    args.MessageToDisplay += exec.ExecutableTitle + "\r\n";
                    //string buf = vFileLocation.Remove(vFileLocation.Length - 4) + "\\" + exec.ExecutablePath;
                    string buf = AppPath + "\\" + exec.ExecutablePath;
                    Process.Start(buf);
                }

                #region Delete temp files
                try
                {
                    File.Delete(vFileLocation);
                    Directory.Delete(vFileLocation.Remove(vFileLocation.Length - 4),true);
                }
                catch (Exception ex)
                {
                    //Files undeletable... It's in temp directory: who cares? ;)
                }
                #endregion

                args.IsUpated = true;
                OnUpdateChanged(args);
            }
            catch (Exception ex)
            {
#if DEBUG
                args.MessageToDisplay = "\r\nImpossible d'exécuter les opérations complémentaires!\r\n" + ex.Message;
#else
                args.MessageToDisplay = "\r\nImpossible d'exécuter les opérations complémentaires!\r\n";
#endif
                args.IsUpated = true;
                OnUpdateChanged(args);
                return;
            }
            #endregion
        }
        #endregion
    }
    /// <summary>
    /// Event args for un update change
    /// </summary>
    public class XmlUpdateEventArgs : EventArgs
    {
        #region Variables
        /// <summary>
        /// Is the update downloading?
        /// </summary>
        public bool IsDownloading { get; set; }
        /// <summary>
        /// Downlaod percentage
        /// </summary>
        public int DownloadPercentage { get; set; }
        /// <summary>
        /// Indicates the message to display
        /// </summary>
        public string MessageToDisplay { get; set; }
        /// <summary>
        /// Indicates whether the update is finished
        /// </summary>
        public bool IsUpated { get; set; }
        #endregion
    }
    /// <summary>
    /// Contains the data of one executable to launch after the update
    /// </summary>
    public struct Executable
    {
        /// <summary>
        /// Relative path of the executable
        /// </summary>
        public string ExecutablePath;
        /// <summary>
        /// Title of the executable or the action it does
        /// </summary>
        public string ExecutableTitle;
    }
}
